#include "../include/mptp.h"
#include "../include/error.h"
#include "../include/url.h"
#include <arpa/inet.h>
#include <errno.h>
#include <netdb.h>
#include <netinet/in.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <strings.h>
#include <sys/socket.h>
#include <unistd.h>

bool lm_mptp_packet_init(lm_mptp_t *packet, bool is_request, uint8_t code, bool is_last) {
  packet->header.flags = 0;
  packet->header.size  = 0;
  bzero(packet->data, MPTP_CHUNK_MAX);

  if (code > MPTP_CODE_MAX) {
    lm_error_set(LM_ERR_MPTPBadCode);
    return false;
  }

  packet->header.flags |= (MPTP_VERSION_SUPPORTED << 8);

  if (is_request)
    packet->header.flags |= (MPTP_REQUEST << 7);
  else
    packet->header.flags |= (MPTP_RESPONSE << 7);

  packet->header.flags |= (code << 4);

  if (is_last || is_request)
    packet->header.flags |= (1 << 3);
  else
    packet->header.flags |= (0 << 3);

  return true;
}

bool lm_mptp_verify(lm_mptp_t *packet){
  if (NULL == packet) {
    lm_error_set(LM_ERR_ArgNULL);
    return false;
  }

  if (MPTP_FLAGS_VERSION(packet) != MPTP_VERSION_SUPPORTED) {
    lm_error_set(LM_ERR_MPTPBadVersion);
    return false;
  }

  if (packet->header.size > MPTP_CHUNK_MAX || packet->header.size < 0) {
    lm_error_set(LM_ERR_MPTPBadChunk);
    return false;
  }

  if (MPTP_FLAGS_CODE(packet) > MPTP_CODE_MAX || MPTP_FLAGS_CODE(packet) < 0) {
    lm_error_set(LM_ERR_MPTPBadCode);
    return false;
  }

  return true;
}

int lm_mptp_socket(char *addr, uint16_t port, struct sockaddr *saddr) {
  if (NULL == addr || NULL == saddr) {
    lm_error_set(LM_ERR_ArgNULL);
    return -1;
  }

  struct addrinfo hints, *res, *cur;
  int             sock = 0, status = 0, family = -1;

  bzero(&hints, sizeof(hints));
  hints.ai_family   = AF_UNSPEC;
  hints.ai_socktype = SOCK_DGRAM;

  if ((status = getaddrinfo(addr, NULL, &hints, &res)) < 0) {
    lm_error_set(LM_ERR_MPTPHostFail);
    return -1;
  }

  for (cur = res; cur != NULL; cur = cur->ai_next) {
    switch (cur->ai_family) {
    case AF_INET:
      family                   = cur->ai_family;
      struct sockaddr_in *ipv4 = (struct sockaddr_in *)cur->ai_addr;
      ipv4->sin_port           = htons(port);

      memcpy(saddr, cur->ai_addr, sizeof(struct sockaddr));
      break;

    case AF_INET6:
      family                    = cur->ai_family;
      struct sockaddr_in6 *ipv6 = (struct sockaddr_in6 *)cur->ai_addr;
      ipv6->sin6_port           = htons(port);

      memcpy(saddr, cur->ai_addr, sizeof(struct sockaddr));
      break;
    }

    if (family != -1)
      break;
  }

  freeaddrinfo(res);

  if (family == -1) {
    lm_error_set(LM_ERR_MPTPHostFail);
    return -1;
  }

  if ((sock = socket(family, SOCK_DGRAM, IPPROTO_UDP)) < 0) {
    lm_error_set(LM_ERR_MPTPSocketFail);
    return -1;
  }

  return sock;
}

void lm_mptp_close(int sock) {
  close(sock);
}

int lm_mptp_client_connect(char *addr, uint16_t port) {
  struct sockaddr saddr;
  int             sock;

  bzero(&saddr, sizeof(saddr));

  if ((sock = lm_mptp_socket(addr, port, &saddr)) < 0)
    return -1;

  struct timeval timeout;
  bzero(&timeout, sizeof(timeout));

  timeout.tv_sec  = MPTP_TIMEOUT;
  timeout.tv_usec = 0;

  if (setsockopt(sock, SOL_SOCKET, SO_RCVTIMEO, (char *)&timeout, sizeof(timeout)) < 0) {
    lm_error_set(LM_ERR_MPTPSetsockopt);
    lm_mptp_close(sock);
    return -1;
  }

  if (connect(sock, &saddr, sizeof(saddr)) < 0) {
    lm_mptp_close(sock);
    lm_error_set(LM_ERR_MPTPConnectFail);
    return -1;
  }

  return sock;
}

bool lm_mptp_client_verify(lm_mptp_t *packet) {
  if (!lm_mptp_verify(packet))
    return false;

  if (MPTP_IS_REQUEST(packet)) {
    lm_error_set(LM_ERR_MPTPNotResponse);
    return false;
  }

  return true;
}

bool lm_mptp_client_send(int sock, lm_mptp_t *packet) {
  if (NULL == packet) {
    lm_error_set(LM_ERR_ArgNULL);
    return false;
  }

  if (MPTP_FLAGS_VERSION(packet) != MPTP_VERSION_SUPPORTED) {
    lm_error_set(LM_ERR_MPTPBadVersion);
    return false;
  }

  if (packet->header.size > MPTP_CHUNK_MAX) {
    lm_error_set(LM_ERR_MPTPBadChunk);
    return false;
  }

  char buffer[sizeof(packet->header) + packet->header.size];
  bzero(buffer, sizeof(buffer));

  if (packet->header.size > 0)
    memcpy(buffer + sizeof(packet->header), packet->data, packet->header.size);

  packet->header.flags = htons(packet->header.flags);
  packet->header.size  = htons(packet->header.size);
  memcpy(buffer, &packet->header, sizeof(packet->header));

  if (send(sock, buffer, sizeof(buffer), 0) < 0) {
    lm_error_set(LM_ERR_MPTPSendFail);
    return false;
  }

  return true;
}

bool lm_mptp_client_recv(int sock, lm_mptp_t *packet) {
  if (NULL == packet) {
    lm_error_set(LM_ERR_ArgNULL);
    return false;
  }

  char buffer[sizeof(packet->header) + MPTP_CHUNK_MAX];
  bzero(buffer, sizeof(buffer));

  if (recv(sock, buffer, sizeof(buffer), 0) < 0) {
    if (ETIMEDOUT == errno || EAGAIN == errno) {
      lm_error_set(LM_ERR_MPTPTimeout);
      return false;
    }
    lm_error_set(LM_ERR_MPTPRecvFail);
    return false;
  }

  memcpy(&packet->header, buffer, sizeof(packet->header));
  packet->header.flags = ntohs(packet->header.flags);
  packet->header.size  = ntohs(packet->header.size);

  if (packet->header.size > MPTP_CHUNK_MAX)
    return true;

  memcpy(packet->data, buffer + sizeof(packet->header), packet->header.size);
  return true;
}

int lm_mptp_server_listen(char *addr, uint16_t port) {
  struct sockaddr saddr;
  int             sock;

  bzero(&saddr, sizeof(saddr));

  if ((sock = lm_mptp_socket(addr, port, &saddr)) < 0)
    return -1;

  if (bind(sock, &saddr, sizeof(struct sockaddr)) < 0) {
    lm_mptp_close(sock);
    lm_error_set(LM_ERR_MPTPBindFail);
    return -1;
  }

  return sock;
}

bool lm_mptp_server_verify(lm_mptp_t *packet) {
  if (!lm_mptp_verify(packet))
    return false;

  if (!MPTP_IS_REQUEST(packet)) {
    lm_error_set(LM_ERR_MPTPNotRequest);
    return false;
  }

  if(!MPTP_IS_LAST(packet)){
    lm_error_set(LM_ERR_MPTPNotLast);
    return false;
  }

  return true;
}

bool lm_mptp_server_recv(int sock, lm_mptp_t *packet, struct sockaddr *addr) {
  if (NULL == packet || NULL == addr) {
    lm_error_set(LM_ERR_ArgNULL);
    return false;
  }

  socklen_t socklen = sizeof(struct sockaddr);
  char      buffer[sizeof(packet->header) + MPTP_CHUNK_MAX];

  bzero(buffer, sizeof(buffer));
  bzero(packet, sizeof(lm_mptp_t));

  if (recvfrom(sock, buffer, sizeof(buffer), 0, addr, &socklen) <= 0) {
    lm_error_set(LM_ERR_MPTPRecvFail);
    return false;
  }

  memcpy(&packet->header, buffer, sizeof(packet->header));
  packet->header.flags = ntohs(packet->header.flags);
  packet->header.size  = ntohs(packet->header.size);

  if (packet->header.size > MPTP_CHUNK_MAX)
    return true;

  memcpy(packet->data, buffer + sizeof(packet->header), sizeof(packet->header));
  return true;
}

bool lm_mptp_server_send(int sock, lm_mptp_t *packet, struct sockaddr *addr) {
  if (NULL == packet) {
    lm_error_set(LM_ERR_ArgNULL);
    return false;
  }

  if (MPTP_FLAGS_VERSION(packet) != MPTP_VERSION_SUPPORTED) {
    lm_error_set(LM_ERR_MPTPBadVersion);
    return false;
  }

  if (packet->header.size > MPTP_CHUNK_MAX) {
    lm_error_set(LM_ERR_MPTPBadChunk);
    return false;
  }

  socklen_t addrlen = sizeof(struct sockaddr);
  char      buffer[sizeof(packet->header.size) + packet->header.size];

  bzero(buffer, sizeof(buffer));

  packet->header.flags = htons(packet->header.flags);
  packet->header.size  = htons(packet->header.size);

  memcpy(buffer, &packet->header, sizeof(packet->header));
  memcpy(buffer + sizeof(packet->header), packet->data, packet->header.size);

  if (sendto(sock, buffer, sizeof(buffer), 0, addr, addrlen) < 0) {
    lm_error_set(LM_ERR_MPTPSendFail);
    return false;
  }

  return true;
}
