#include "../include/pool.h"
#include "../include/error.h"
#include "../include/mptp.h"
#include "../include/util.h"
#include <sys/socket.h>
#include <stdlib.h>
#include <stdio.h>

void lm_pool_free(lm_pool_t *pool) {
  lm_url_free(&pool->url);
  free(pool);
}

lm_pool_t *lm_pool_new(char *name, char *url) {
  lm_pool_t *pool = malloc(sizeof(lm_pool_t));
  pool->available = true;
  pool->name      = name;

  if (!lm_url_init(&pool->url, url)) {
    free(pool);
    return NULL;
  }

  if(pool->url.empty)
    return pool;

  if (!eq(pool->url.protocol, "mptp")) {
    lm_error_set(LM_ERR_PoolNoSupport);
    lm_pool_free(pool);
    return NULL;
  }

  return pool;
}

void lm_pool_test(lm_pool_t *pool) {
  if(pool->url.empty){
    pool->available = false;
    return;
  }

  lm_mptp_t packet;
  lm_mptp_packet_init(&packet, true, MPTP_C2S_PING, true);

  int sock = lm_mptp_client_connect(pool->url.host, pool->url.port);
  if (sock == -1) {
    pool->available = false;
    return;
  }

  if (!lm_mptp_client_send(sock, &packet)) {
    pool->available = false;
    goto end;
  }

  if (!lm_mptp_client_recv(sock, &packet)) {
    pool->available = false;
    goto end;
  }

  if (!lm_mptp_client_verify(&packet)){
    pool->available = false;
    goto end;
  }

  pool->available = MPTP_FLAGS_CODE(&packet) == MPTP_S2C_PONG;
end:
  lm_mptp_close(sock);
  return;
}

bool lm_ctx_pool_add(lm_ctx_t *ctx, char *name, char *url) {
  lm_pool_t *pool = lm_pool_new(name, url);
  if (NULL == pool)
    return false;

  pdebug(ctx, __func__, "pool name is %s", pool->name);
  if(!pool->url.empty)
    pdebug(ctx, __func__, "pool URL is %s://%s:%d%s", pool->url.protocol, pool->url.host, pool->url.port, pool->url.path);

  if (NULL == ctx->pools) {
    ctx->pools = pool;
    return true;
  }

  lm_pool_t *cur = ctx->pools;
  while (NULL != cur) {
    if (NULL == cur->next) {
      cur->next = pool;
      return true;
    }
    cur = cur->next;
  }

  return false;
}

bool lm_ctx_pool_del(lm_ctx_t *ctx, char *name) {
  lm_pool_t *cur = ctx->pools, *prev = NULL;
  while (NULL != cur) {
    if (!eq(cur->name, name)) {
      cur = cur->next;
      continue;
    }

    if (NULL == prev) {
      ctx->pools = cur->next;
      lm_pool_free(cur);
      return true;
    }

    prev->next = cur->next;
    lm_pool_free(cur);
    return true;
  }

  return false;
}

void lm_ctx_pool_clear(lm_ctx_t *ctx) {
  lm_pool_t *cur = ctx->pools, *prev = NULL;
  while (NULL != cur) {
    prev = cur;
    cur  = cur->next;
    lm_pool_free(prev);
  }

  ctx->pools = NULL;
}

void lm_ctx_pool_test(lm_ctx_t *ctx) {
  lm_pool_t *cur = ctx->pools;
  while (NULL != cur) {
    lm_pool_test(cur);
    if (!cur->available)
      pdebug(ctx, __func__, "%s is not avaliable: %s", cur->name, lm_strerror());
    cur = cur->next;
  }
}

bool lm_ctx_pool_serve(lm_ctx_t *ctx, char *addr) {
  struct sockaddr saddr;
  char *host = NULL;
  lm_mptp_t packet;
  uint16_t port;
  int sock;

  if(!parse_host(addr, host, &port))
    return false;

  if(port == 0){
    lm_error_set(LM_ERR_NoPort);
    return false;
  }

  if((sock = lm_mptp_server_listen(addr, port)) < 0)
    return false;

  while(lm_mptp_server_recv(sock, &packet, &saddr)){
    if(!lm_mptp_verify(&packet)){
      pdebug(ctx, __func__, "skipping invalid packet: %s", lm_strerror());
      continue;
    }

    switch (MPTP_FLAGS_CODE(&packet)) {
    case MPTP_C2S_PING:
      lm_mptp_packet_init(&packet, false, MPTP_S2C_PONG, true);
      lm_mptp_server_send(sock, &packet, &saddr);
      break;
    }
  }

  return true;
}
